"use strict";
this.name        = "stardestroyer-fighter";
this.author      = "Norby";
this.copyright   = "2016 Norbert Nagy";
this.licence     = "CC BY-NC-SA 4.0";
this.description = "support functions for fighters";

this.$FighterHull = this.ship.maxEnergy; //must repair in hangar, avoid recharge when no shields
this.$FighterNumber = 0; //order of this fighter in escort formation
this.$FighterOwner = null; //store the mothership
this.$FighterShields = false; //cache of this.ship.scriptInfo.npc_shields
this.$OwnerScoopPos = null; //store the mothership's scoop position for landing
this.$TimerLandNow = null; //trigger _landNow() after some time for sure landing
this.$ws = worldScripts.stardestroyer;

//ship script events
this.shipAttackedOther = function(other) {
        if( other && other.isDerelict && other == this.ship.target )
                this.ship.target = null; //do not destroy derelicts, choose another target instead
}

this.shipBountyChanged = function(delta,reason) {
        //prevent orange scanner flag on player's fighters with ShipVersion OXP
        if(this.$FighterOwner == player.ship && this.ship.bounty > 0 ) {
                this.ship.bounty = 0; //Warning: this change trigger this event again!
        }                             //So must check before that bounty is greater than 0
}                                     //else this is an infinite loop which make Crash To Desktop!

this.shipCollided = this.shipWasScooped = function(otherShip) {
        if(otherShip == this.$FighterOwner) {
            this.ship.energy = this.ship.maxEnergy; //sometimes prevent destruction

            if(this.ship.AIState == "LANDING") //put into the hangar
                this._landNow(this.ship, this.$FighterOwner);
        }
}

this.shipTakingDamage = function(amount, whom, type) {
        this.$FighterHull -= amount; //workaround recharge when no shields
        if( !this.$FighterShields ) {
                if(this.$FighterHull < 0) {
                    delete this.shipTakingDamage;//prevent infinite loop
                    this.ship.explode();
                }
        }
}

this.shipSpawned = function() {
        if( this.ship && this.ship.scriptInfo && this.ship.scriptInfo.npc_shields == "no" ) {
                this.$FighterShields = false;
        } else this.$FighterShields = true;
}

this.spawnedAsEscort = function(mother) {  //for NPC mothers
        if(mother && mother.script) {
            if(!mother.script.$LaunchedFighters) mother.script.$LaunchedFighters = 1;
            else mother.script.$LaunchedFighters++;
            this.$FighterNumber = mother.script.$LaunchedFighters;
            if(!mother.script.$Fighters) mother.script.$Fighters = [];
            mother.script.$Fighters.push(this.ship);
            this.ship.displayName = this.ship.name+" #"+this.$FighterNumber;
        }
        this.$FighterOwner = mother;
}

this.shipTargetDestroyed = function(target) {
        if(this.ship == player.ship || this.$FighterOwner != player.ship )
            return; //exit if worldScript, do in ship script only
        if(target.primaryRole == "constrictor" && missionVariables.conhunt && missionVariables.conhunt == "STAGE_1") {
                // just in case an escort kills the constrictor, let's not break the mission for the player...
                missionVariables.conhunt = "CONSTRICTOR_DESTROYED";
        }
        
        if(target.isRock || target.isBoulder || target.isCargo || target.isWeapon)
                return;
        
//      player.score += 1; - score should remain personal
        var b = "";
        if(target.bounty > 0) {
            b = " : " + target.bounty + "₢ awarded.";
            player.credits += target.bounty;
        }
        player.consoleMessage(this.ship.displayName+" killed "+target.name+b, 5);
        log(this.name, this.ship.displayName+" killed " + target.name + " : " + target.bounty +"₢");
}

this.shipDied = function(whom,why) {
        if(this.ship == player.ship) return; //exit if worldScript, do in ship script only
        //var w = worldScripts.escortdeck;

        //increase steering and acceleration in proportion with the smaller mass on deck
        //w.$EscortDeck_SetMaxs( player.ship, w );

        //var pad = w.$EscortDeckShip.indexOf(this.ship);
        //if( pad > -1 ) w.$EscortDeckShipData[pad] = null; //prevent recreation in dock
        
        if( whom && whom.isValid && whom != player.ship && this.$FighterOwner == player.ship ) {
                player.commsMessage("A "+this.ship.name+" is lost.", 5);
                this.$ws._MFD(); //update numbers in MFD
        }
        log(this.name, this.ship.displayName+" terminated by "+whom+" "+why);
}


//fighter functions

this._escortPosition = function() {this._escortPosition2(this.ship);} //for stardestroyer-fighterAI.plist

this._escortPosition2 = function(ship) {
        var sc = ship.script;
        if( !sc || !sc.$FighterNumber || !sc.$FighterOwner || !sc.$FighterOwner.isValid
            || ship == player.ship ) return; //for sure
        ship.scanClass = "CLASS_BUOY";//for scoop deactivation at stopped landing and no masslock

        if( sc.$FighterOwner == player.ship && player.alertCondition == 1 ) { //green alert
            ship.AIState == "LANDING";
            this.$ws._MFD(); //update if owner is the player
            return;
        }

        if( ship.target && ship.target.hasHostileTarget ) {
            ship.setAI("interceptAI.plist");
            return;
        }

        var epos = sc.$ws.coordinatesForEscortPosition(sc.$FighterNumber);
        var pos = sc.$FighterOwner.position.add(epos);
        ship.savedCoordinates = pos;
        if( ship.injectorSpeedFactor > 0 )
            ship.desiredSpeed = ship.injectorSpeedFactor * ship.maxSpeed;
        else ship.desiredSpeed = 7 * ship.maxSpeed; //for Oolite 1.80
        ship.desiredRange = 0; //go to the escort position in the formation
        ship.performFlyToRangeFromDestination();
        ship.reactToAIMessage("OWNER_FAR");//go back to owner
        //var dist = ship.position.distanceTo(sc.$FighterOwner.position);
        //if(dist > 10000) ship.reactToAIMessage("OWNER_FAR");//go back to owner
        //else if(dist > 5000) ship.reactToAIMessage("OWNER_NEAR");//go back to owner
        //else ship.reactToAIMessage("OWNER_NEXT");//keep formation
}

this._landing = function() {this._landing2(this.ship);}

this._landing2 = function(ship) {
        var sc = ship.script;
        if( !sc || ship == player.ship ) return; //for sure
        ship.scanClass = "CLASS_CARGO";//for scoop activation

        var owner = sc.$FighterOwner;
        if( !owner || !owner.isValid ) return;

        ship.target = null; //clear target to stop attack

        var landingdist = 50; //m from the scoop position of owner where put into hangar
        //make a waypoint below the owner's scoop position
        var scooppos = owner.position;
        var sp = this.$OwnerScoopPos;
        if( !sp ) { //first time read scoop pos from shipdata.plist
            var shipdata = Ship.shipDataForKey(owner.dataKey);
            if( shipdata ) {
                sp = shipdata["scoop_position"];
                if(sp && sp != "undefined") {
                    sp = sp.split(" ",3);
                    if( sp && sp.length == 3 ) {
                        this.$OwnerScoopPos = sp;
                        //log(this.name, owner.name+" scoop_position:"+sp); //debug
                    }
                }
            }
            if( !this.$OwnerScoopPos ) { //no scoop pos in shipdata so make one
                this.$OwnerScoopPos = [];
                this.$OwnerScoopPos[0] = 0;
                this.$OwnerScoopPos[1] = -20;
                this.$OwnerScoopPos[2] = 0;
                sp = this.$OwnerScoopPos;
            }
        }
        scooppos = scooppos.add(owner.vectorRight.multiply(sp[0]))
                            .add(owner.vectorUp.multiply(sp[1]))
                            .add(owner.heading.multiply(sp[2]));
        var down = 1.5*owner.collisionRadius; //AI never enter into collisionRadius
        var downpos = scooppos.add(owner.vectorUp.multiply(-down));
        ship.savedCoordinates = downpos.add(owner.velocity.multiply(6)); //go below the owner
        // also must use setDestinationFromCoordinates in AI.plist

        var d = ship.position.distanceTo(scooppos);
        var d2 = ship.position.distanceTo(downpos);
        //log(this.name, d);
        if( this.$ws.$TractorBeam ) {
                if( d < landingdist ) { //enough near?: add to the hangar
                        this._landNow(ship, owner);
                        return;
                }
                if( d2 < down ) { //arrived below the owner?
                        //if( owner.addCollisionException ) { //need Oolite v1.81
                        //    owner.addCollisionException(ship); //must go near
                            //log(this.name, "coll.ex.len:"+owner.collisionExceptions.length);//debug
                        //}

                        //go up to the owner's center point from below into the dock
                        ship.savedCoordinates = scooppos.add(owner.velocity.multiply(2));
                        // also must use setDestinationFromCoordinates in AI.plist

                        //apply "tractor beam" to pull this fighter to the scoop
                        var v = scooppos.subtract(ship.position);
                        ship.velocity = v.direction().multiply(200).add(v.multiply(0.4))
                                .add(owner.velocity);

                        //make sure the fighter will land after 10 seconds
                        if( !this.$TimerLandNow )
                                this.$TimerLandNow = new Timer(this, this._TimedLandNow.bind(this), 10);

                        ship.desiredSpeed = 0; //must set before "OWNER_NEXT"
                        ship.reactToAIMessage("OWNER_NEXT"); //go straight
                        return;
                }
        } else { //land without tractor beam
                if( d < landingdist ) {
                        //must be triggered outside collisionRadius of owner
                        this._landNow(ship, owner); //add to the hangar
                        return;
                } else if( d2 < down ) { //surely land after 10 seconds
                        if( !this.$TimerLandNow )
                                this.$TimerLandNow = new Timer(this, this._TimedLandNow.bind(this), 10);
                        //go up to the owner's center point from below into the dock
                        ship.savedCoordinates = scooppos.add(owner.velocity);
                        ship.desiredSpeed = ship.maxSpeed; //must set before "OWNER_NEXT"
                        ship.reactToAIMessage("OWNER_NEXT"); //go straight
                        return;
                }
        }

        if( d < 2000 ) ship.reactToAIMessage("OWNER_NEAR"); //approach owner at normal speed
        else ship.reactToAIMessage("OWNER_FAR"); //use injectors in approach if available
}

this._landNow = function(ship, owner) { //put back into the internal hangar
        if(!ship || !owner || !owner.isValid) return; //for sure
        var f = [], f2 = []; //remove from $Fighters array
        if( owner == player.ship ) {
            if( this.$ws.$LaunchedFighters > 0 ) this.$ws.$LaunchedFighters--;
            f = this.$ws.$Fighters;
        } else if( owner.script ) {
            if( owner.script.$LaunchedFighters > 0 ) owner.script.$LaunchedFighters--;
            f = owner.script.$Fighters;
        }
        var out = 0;
        for(var i = 0; i < f.length; i++) {
            var fi = f[i];
            if( fi != ship ) {
                f2.push(fi);
                if( fi && fi.isValid ) out++;
            }
        }
        if( owner == player.ship ) {
            this.$ws.$Fighters = f2;
            this.$ws._MFD(); //update numbers in MFD
            if( out == 0 ) {
                player.commsMessage("All fighters landed", 10);
                player.consoleMessage(" ", 10); //make room to the countdown
            }
        } else if( owner.script ) owner.script.$Fighters = f2;

        //prevent double landing if both shipCollided and shipWasScooped is fired
        var sc = ship.script;  
        if( sc ) sc.$FighterOwner = null;

        //if( owner.removeCollisionException ) { //need Oolite v1.81
        //    owner.removeCollisionException(ship); //must go near
            //log(this.name, "coll.ex.len:"+owner.collisionExceptions.length);//debug
        //}
        ship.remove(true); //landed
}

this._newTarget = function() {this._newTarget2(this.ship);} //AI in plist located new target

this._newTarget2 = function() { //AI in plist located new target
        var sc = ship.script;
        if( !sc || ship == player.ship ) return; //for sure
        var owner = sc.$FighterOwner;
        if( !owner || !owner.isValid || owner != player.ship ) return; //player's fighters only
        var t = ship.target;
        if( !this.$ws._isValidFighterTarget(t, owner) ) return;

        player.consoleMessage(ship.name+" #"+this.$FighterNumber+" aim "+t.displayName, 4.5);
        this.$ws._MFD(); //update numbers in MFD
}

this._TimedLandNow = function() { 
        if( this.$TimerLandNow ) {
                this.$TimerLandNow.stop();
                delete this.$TimerLandNow;
        }
        this._landNow(this.ship, this.$FighterOwner);
}
